#2.0

import json
import os
import re
from . import core
from .__init__ import MAIN_PATH
from .app_utils import urlencoded_to_dict
from .http_parser import Redirect
from .url_utils import UrlTemplate, split_route_scope, format_ip_port 
from .http_parser import PartialContent
from .http_parser import HTTP_Status_Code
from .special_media_type import Response
from .http_parser import HTTP_Message_Header_Line as Header_Line
from .http_parser import HTTP_Message_Request_Header_Tag as Request_Header_Tag
from .http_parser import HTTP_Message_Response_Header_Tag as Response_Header_Tag
from .metrics import get_timeout_metrics

class MIME_TYPE:
    HTML = 'text/html'
    JAVA_SCRIPT = 'text/javascript'
    CSS = 'text/css'
    ICO = "image/x-icon"
    PNG = "image/png"
    SVG = "image/svg+xml"
    TEXT = "text/plain"
    MP4 = "video/mp4"
    JSON = "application/json"
    WEBM_AUDIO = "audio/webm"


class STORE_VARS:
    #TODO: UPDATE Vars
    '''Keys used to access attributes of the dict provided to the respective handler methods'''

    COOKIES = 'cookies'
    QUERY_STRING = 'query_string'
    RESPONSE = 'response'
    POST = 'post'
    FLAGS = 'flags'
    REQUEST_HEADER = 'request_header'
    TEMPLATE_VARS = 'template_args'
    

class LOGGING_OPTIONS:

    '''Options used to set the logging mode via the ``set_logging`` function'''

    REQUESTS = 'request'
    RESPONSES = 'response'
    DEBUG = 'debug'
    INFO = 'info'
    WARNING = 'warning'
    ERROR = 'error'
    CRITICAL = 'critical'
    TIME = 'time'
    TIMEOUT = 'timeout'


def _scope_value(value):
    return value if value is not None else '*'


def _store_static_route(method, route, func, mime_type):
    ip, port, domain, path = split_route_scope(route)
    ip_key = _scope_value(ip)
    port_key = _scope_value(port)
    domain_key = _scope_value(domain)
    static_routes = core.ROUTES[method]['static']
    key = (ip_key, port_key, domain_key)
    if key not in static_routes:
        static_routes[key] = {}
    static_routes[key][path] = {
        'handler': func,
        'type': mime_type,
    }


def _template_weight(entry):
    weight = 0
    if entry.get('ip') is not None:
        weight += 4
    if entry.get('port') is not None:
        weight += 2
    if entry.get('domain') is not None:
        weight += 1
    return weight


def _register_template_route(method, template_obj, func, mime_type):
    entry = {
        'ip': template_obj.ip,
        'port': template_obj.port,
        'domain': template_obj.domain,
        'template': template_obj,
        'handler': func,
        'type': mime_type,
    }
    templates = core.ROUTES[method]['templates']
    templates = [item for item in templates if item['template'] != template_obj]
    templates.append(entry)
    templates.sort(key=_template_weight, reverse=True)
    core.ROUTES[method]['templates'] = templates


def _parse_scope(scope):
    if isinstance(scope, tuple):
        ip = scope[0] if len(scope) > 0 else None
        port = scope[1] if len(scope) > 1 else None
        domain = scope[2] if len(scope) > 2 else None
        if isinstance(port, str) and port.isdigit():
            port = int(port)
        return (ip or None, port, domain or None)
    if not scope:
        return (None, None, None)
    fake_route = scope if scope.endswith('/') else f'{scope}/'
    ip, port, domain, _ = split_route_scope(fake_route)
    if isinstance(port, str) and port.isdigit():
        port = int(port)
    return (ip, port, domain)


def _scope_repr(ip, port, domain):
    if ip is None and port is None and domain is None:
        return '*'
    ip_part = format_ip_port(ip, port) or '*'
    domain_part = domain if domain is not None else '*'
    return f'{ip_part}::{domain_part}'

def _register_method_route(method, route, mime_type, func, handler_map, template_list, acceptable_arg_counts=(1,)):
    if acceptable_arg_counts is not None:
        argcount = func.__code__.co_argcount
        if argcount not in acceptable_arg_counts:
            allowed = ', '.join(str(c) for c in acceptable_arg_counts)
            raise Exception(f'The function decorated for method "{method}" must accept {allowed} argument(s).')
    if isinstance(route, str):
        handler_map[route] = [func, mime_type]
        _store_static_route(method, route, func, mime_type)
    elif isinstance(route, UrlTemplate):
        route.type = mime_type
        route.handler = func
        _register_template_route(method, route, func, mime_type)
        template_list[:] = [template for template in template_list if template != route]
        template_list.append(route)
    else:
        raise Exception('Unsupported route type.')
    return func

def register(*args, **kwargs):
    """
    Decorator to register a GET route. Accepts either raw strings or `UrlTemplate`
    instances using the `ip:port::domain:/path` selector syntax.

    Examples:
        - `register(route='::/status', type=MIME_TYPE.JSON)`              -> matches all endpoints
        - `register(route='127.0.0.1:8443::/admin', type=MIME_TYPE.HTML)` -> specific IP/port
        - `register(route=':::example.com:/only-domain', ...)`            -> any IP/port, domain `example.com`
        - `register(route=':8080::example.com:/info', ...)`               -> any IP, port 8080, domain `example.com`
        - `register(route=UrlTemplate('[::1]:8000::/v1/{resource:str}'), ...)` -> IPv6 template with placeholders

    When the handler expects an argument it receives the session dict. Supported return
    values include `str`, `bytes`, `Dynamic`, `PartialContent`, `Redirect`, or `Response`.
    """

    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    mime_type = kwargs['type']
    def inner(func):
        if isinstance(route, str):
            core.PAGES[route] = [func, mime_type]
            _store_static_route('GET', route, func, mime_type)
        elif isinstance(route, UrlTemplate):
            route.type = mime_type
            route.handler = func
            _register_template_route('GET', route, func, mime_type)
            core.GET_TEMPLATES[:] = [template for template in core.GET_TEMPLATES if template != route]
            core.GET_TEMPLATES.append(route)
        else:
            raise Exception('Unsupported route type.')
        return func            
    return inner

def post_handler(*args, **kwargs):

    '''
    ``General Information:``\n
    The post_handler decorator adds the decorated function to the servers POST_HANDLER
    dictionary. The decorater requires a path and a type argument to be specified the decorated function
    gets called whenever a post is targeted to the speciefied path. The function is expected to return a string of the
    under the type attribute declared MIME-type. The function decorated gets passed a dictionary containing
    the variables passed with the ``POST so it has to take an argument``
    
    ``Usage:``
    ```python
    @post_handler(route= '/', type= MIME_TYPE.TEXT)
    def main(args):
        return str(args)
    ```
    ``Note:`` The ``args`` argument is necessary!
    '''
    

    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    mime_type = kwargs['type']
    def inner(func):
        return _register_method_route('POST', route, mime_type, func, core.POST_HANDLER, core.POST_TEMPLATES, acceptable_arg_counts=(1,))
    return inner

def put_handler(*args, **kwargs):
    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    mime_type = kwargs['type']
    def inner(func):
        return _register_method_route('PUT', route, mime_type, func, core.PUT_HANDLER, core.PUT_TEMPLATES, acceptable_arg_counts=(1,))
    return inner

def delete_handler(*args, **kwargs):
    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    mime_type = kwargs['type']
    def inner(func):
        return _register_method_route('DELETE', route, mime_type, func, core.DELETE_HANDLER, core.DELETE_TEMPLATES, acceptable_arg_counts=(0,1))
    return inner

def patch_handler(*args, **kwargs):
    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    mime_type = kwargs['type']
    def inner(func):
        return _register_method_route('PATCH', route, mime_type, func, core.PATCH_HANDLER, core.PATCH_TEMPLATES, acceptable_arg_counts=(1,))
    return inner

def options_handler(*args, **kwargs):
    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    mime_type = kwargs['type']
    def inner(func):
        return _register_method_route('OPTIONS', route, mime_type, func, core.OPTIONS_HANDLER, core.OPTIONS_TEMPLATES, acceptable_arg_counts=(0,1))
    return inner


def error_handler(*args, **kwargs):

    '''
    ``General Information:``\n
    The error_handler decorator is used to serve a specific page if an error-code occurs.
    ``
    
    ``Usage:``
    ```python
    @error_handler(error_code= 404, type= MIME_TYPE.TEXT)
    def main():
        return str('Page not found!')
    ```
    '''
    

    if 'error_code' not in kwargs:
        raise Exception('The "error_code" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    err = kwargs['error_code']
    def inner(func):
        if func.__code__.co_argcount != 0:
            raise Exception('The function decorated with the error_handler is not allowed to take arguments.')
        if kwargs['error_code'] not in core.ERROR_HANDLER:
                core.ERROR_HANDLER[err] = [func,kwargs['type']]
        return func            
    return inner

def load_file(path:str) -> str:

    '''The load_file function attempts to read the content of a file with given path and returns it as a utf-8 encoded string'''

    with open(MAIN_PATH+path,'r',encoding='utf-8') as content:
            lines = content.readlines()
            return ''.join(lines)

def load_bin_file(path:str) -> bytes:

    '''The load_file function attempts to read the content of a file with given path and returns it as bytes'''

    path = MAIN_PATH+path
    size = os.path.getsize(path)
    with open(path,'rb') as content:
            return content.read(size)

#TODO: File through new file not found error so status code 404 can be returned
def load_file_from_directory(root_path: str, file_path: str) -> str:
    """The load_file_from_directory function attempts to read the content of a file with given path 
    relative to root_path and returns it as a utf-8 encoded string.
    """
    
    abs_root_path = os.path.abspath(root_path)
    abs_file_path = os.path.abspath(os.path.join(root_path, file_path))
    if os.path.commonpath([abs_root_path]) != os.path.commonpath([abs_root_path, abs_file_path]):
        raise ValueError("Attempt to access file outside the root directory")
    
    with open(abs_file_path, 'r', encoding='utf-8') as content:
        return content.read()

def load_bin_file_from_directory(root_path: str, file_path: str) -> bytes:
    """The load_bin_file_from_directory function attempts to read the content of a file with given path 
    relative to root_path and returns it as bytes.
    """
    
    abs_root_path = os.path.abspath(root_path)
    abs_file_path = os.path.abspath(os.path.join(root_path, file_path))
    

    if os.path.commonpath([abs_root_path]) != os.path.commonpath([abs_root_path, abs_file_path]):
        raise ValueError("Attempt to access file outside the root directory")
    
    with open(abs_file_path, 'rb') as content:
        return content.read()



def render_page(path:str, args:dict) -> str:

    '''The render_page function reads a files content and executes python-code wrapped in double curly braces "{{".
    If a variable name starts with an underscore "_" the python code wrapped in curley braces will be substituded by it's content.
    The values passed with the args dictionary can be accessed with the globals() function in the targeted file'''

    content = ''
    try:
        with open(MAIN_PATH+path,'r',encoding='utf-8') as page:
            content = '\n'.join(page.readlines())
            res_content = content[:]
        to_eval = re.finditer("{{(.|\n)*?}}",content)
        for var in to_eval:
            OUTPUT = {}
            var = content[var.span()[0]:var.span()[1]]
            exec(compile(var.strip('{{}}'),'temp','exec'),args,OUTPUT)
            insert = '\n'.join([str(OUTPUT[x]) for x in OUTPUT.keys() if x[0] == '_'])
            res_content = res_content.replace(var,insert)
    except Exception as e:
        print('[RENDERER] Error with rendering Page.')
        print(e)
    return res_content

def substitude_vars(content:str, vars:dict) -> str:
    
    '''This function accepts a string and substitudes all "%% + key + %%" with the according value from the dictionary'''
    
    try:
        _content = content
        for key in vars:
            _content = re.sub(f'%%{key}%%',vars[key],_content)
    except:
        print('[SUBSTITUDE_VARS] Error substituding vars.')
    return _content

def set_cookie(args, key, value) -> bool:
    if key not in args[STORE_VARS.COOKIES]:
        res:Response = args[STORE_VARS.RESPONSE]
        res.header.add_header_line(Header_Line(Response_Header_Tag.SET_COOKIE,f'{key}={value}'))
        return True
    return False


def is_partial(args) -> bool:
    return 'partial' in args[STORE_VARS.FLAGS]


def set_logging(option:str, state:bool, scope=None) -> None:
    if option not in core.LOGGING_OPTIONS:
        print(f'[LOGGING] Error: There is no logging-option called "{option}".')
        return
    if scope is None:
        core.LOGGING_OPTIONS[option] = state
        print(f'[LOGGING] logging-option "{option}" set to {state}.')
        return
    ip, port, domain = _parse_scope(scope)
    scope_key = (_scope_value(ip), _scope_value(port), _scope_value(domain))
    scoped_options = core.LOGGING_SCOPED_OPTIONS.get(scope_key)
    if not scoped_options:
        scoped_options = dict(core.LOGGING_OPTIONS)
        core.LOGGING_SCOPED_OPTIONS[scope_key] = scoped_options
    scoped_options[option] = state
    print(f'[LOGGING] logging-option "{option}" set to {state} for scope {_scope_repr(ip, port, domain)}.')

def set_logging_callback(callback, scope=None):
    try:
        assert callable(callback), 'The logging callback is not a callable'
        assert callback.__code__.co_argcount == 3, 'The logging callback needs to accept 3 arguments (message, time_stamp, loglvl)'
        if scope is None:
            core.LOGGING_CALLBACK.append(callback)
        else:
            ip, port, domain = _parse_scope(scope)
            scope_key = (_scope_value(ip), _scope_value(port), _scope_value(domain))
            core.LOGGING_SCOPED_CALLBACKS.append({'scope': scope_key, 'callback': callback})
        print('[LOGGING]', "Logging callback has been set")
    except Exception as e:
        print('[LOGGING]','Error:',e)

def log_to_file(path='/log.txt', logging_options=[LOGGING_OPTIONS.DEBUG], scope=None):
    try:
        for option in logging_options:
            assert option in core.LOGGING_OPTIONS, f'log_to_file: There is no logging-option called "{option}".'

        def file_writer_callback(message, time_stamp, loglvl):
            if  loglvl in logging_options:
                with open(MAIN_PATH+path,'a') as file:
                    file.write(f'({time_stamp}) {message}\n')
        if scope is None:
            core.LOGGING_CALLBACK.append(file_writer_callback)
        else:
            ip, port, domain = _parse_scope(scope)
            scope_key = (_scope_value(ip), _scope_value(port), _scope_value(domain))
            core.LOGGING_SCOPED_CALLBACKS.append({'scope': scope_key, 'callback': file_writer_callback})
        print('[LOGGING]', f"File-logging has been activated path= {path}")
        
    except Exception as e:
        print('[LOGGING]', e)


_TIMEOUT_METRICS_ROUTE = None


def expose_timeout_metrics(route: str = '::/_bbws/metrics/timeouts') -> str:
    """
    Register (or move) the built-in timeout metrics endpoint.

    Returns the route selector string that now exposes the endpoint.
    """
    global _TIMEOUT_METRICS_ROUTE

    if _TIMEOUT_METRICS_ROUTE == route:
        return route

    @register(route=route, type=MIME_TYPE.JSON)
    def _timeout_metrics_endpoint():
        return json.dumps(get_timeout_metrics())

    _TIMEOUT_METRICS_ROUTE = route
    return route


# Register the default timeout metrics endpoint so operators can scrape it immediately.
expose_timeout_metrics()
        

def enable_cors(
    allow_origin='*',
    allow_methods=None,
    allow_headers=None,
    expose_headers=None,
    allow_credentials=False,
    max_age=600,
) -> None:
    if allow_methods is None:
        allow_methods = ['GET', 'POST', 'OPTIONS']
    if allow_headers is None:
        allow_headers = ['*']
    if expose_headers is None:
        expose_headers = []

    def _normalize(value, transform=None):
        if isinstance(value, (list, tuple, set)):
            items = list(value)
        else:
            items = [value]
        if transform:
            return [transform(item) for item in items]
        return [str(item) for item in items]

    origin_value = allow_origin
    if isinstance(allow_origin, (list, tuple, set)):
        origin_value = list(allow_origin)

    if allow_methods == '*' or (isinstance(allow_methods, str) and allow_methods.strip() == '*'):
        normalized_methods = ['*']
    else:
        normalized_methods = _normalize(allow_methods, lambda item: str(item).upper())
        if 'OPTIONS' not in normalized_methods:
            normalized_methods.append('OPTIONS')

    if allow_headers == '*' or (isinstance(allow_headers, str) and allow_headers.strip() == '*'):
        normalized_headers = ['*']
    else:
        normalized_headers = _normalize(allow_headers)

    normalized_expose = _normalize(expose_headers)

    core.CORS_SETTINGS.update({
        'enabled': True,
        'allow_origin': origin_value,
        'allow_methods': normalized_methods,
        'allow_headers': normalized_headers,
        'expose_headers': normalized_expose,
        'allow_credentials': bool(allow_credentials),
        'max_age': int(max_age),
    })
    print('[CORS]', 'CORS support enabled.')


def disable_cors() -> None:
    core.CORS_SETTINGS['enabled'] = False
    core.CORS_SETTINGS['expose_headers'] = []
    print('[CORS]', 'CORS support disabled.')


def get_cors_settings() -> dict:
    return dict(core.CORS_SETTINGS)


def response(content=None, status=None, headers=None, mime_type=None) -> Response:
    return Response(content=content, status=status, headers=headers, mime_type=mime_type)


def server_task(task, interval):
    if not callable(task):
        raise Exception('server_task expects a callable task.')
    if interval < 0:
        raise Exception('server_task expects a non-negative interval.')
    return core.schedule_task(task, interval)

 
def get_pages() -> dict:
    return core.PAGES

def get_sessions() -> dict:
    return core.SESSIONS
    
def get_post_handler() -> dict:
    return core.POST_HANDLER

def start() -> None:

    '''The start function causes the server to start by invoking core.start()'''

    core.start()




from __future__ import annotations

import itertools
import os
import queue
import socket
import ssl
import threading
import time
from dataclasses import dataclass
from multiprocessing import get_context, connection as mp_connection
from multiprocessing.connection import Connection
from multiprocessing.reduction import send_handle, recv_handle
from typing import Any, Dict, Optional

from . import metrics as metrics_module


@dataclass
class TaskRecord:
    task_id: int
    worker_id: int
    instance: Any
    handle: dict
    handler_timeout: Optional[float]


class ProcessWorker:
    def __init__(self, ctx, worker_id: int, snapshot, shared_resources):
        self.ctx = ctx
        self.worker_id = worker_id
        self.snapshot = snapshot
        self.shared_resources = shared_resources
        self.parent_conn: Connection
        self.child_conn: Connection
        self.parent_conn, self.child_conn = ctx.Pipe()
        self.process = None

    def start(self):
        if self.process and self.process.is_alive():
            return
        self.process = self.ctx.Process(
            target=_worker_main,
            args=(self.worker_id, self.child_conn, self.snapshot, self.shared_resources),
            daemon=True,
        )
        self.process.start()

    def restart(self):
        self.stop()
        self.parent_conn, self.child_conn = self.ctx.Pipe()
        self.start()

    def stop(self):
        if self.process and self.process.is_alive():
            try:
                self.parent_conn.send({'cmd': 'STOP'})
            except Exception:
                pass
            self.process.join(timeout=1)
        if self.process and self.process.is_alive():
            self.process.terminate()
            self.process.join(timeout=1)
        self.process = None

    def send_task(self, task_id: int, conn: socket.socket, addr, server_payload):
        if not self.process or not self.process.is_alive():
            raise RuntimeError('Worker process not available.')
        message = {
            'cmd': 'START',
            'task_id': task_id,
            'addr': addr,
            'server': server_payload,
            'socket_info': {
                'family': conn.family,
                'type': conn.type,
                'proto': conn.proto,
            }
        }
        self.parent_conn.send(message)
        send_handle(self.parent_conn, conn.fileno(), self.process.pid)


class ProcessWorkerPool:
    def __init__(self, config, snapshot_builder, resource_exporter):
        self.config = config
        self.snapshot_builder = snapshot_builder
        self.resource_exporter = resource_exporter
        self.ctx = get_context('spawn')
        self.worker_count = max(1, getattr(config, 'WORKER_PROCESSES', 1))
        self.workers: list[ProcessWorker] = []
        self.available = queue.Queue()
        self.task_seq = itertools.count(1)
        self.tasks: Dict[int, TaskRecord] = {}
        self.lock = threading.Lock()
        self.running = False
        self.listener_thread = None
        self.listener_stop = threading.Event()
        self.conn_map: Dict[Connection, int] = {}

    def start(self):
        if self.running:
            return
        snapshot = self.snapshot_builder()
        shared_resources = self.resource_exporter()
        self.available = queue.Queue()
        self.conn_map = {}
        with self.lock:
            self.tasks.clear()
        self.workers = []
        for worker_id in range(self.worker_count):
            worker = ProcessWorker(self.ctx, worker_id, snapshot, shared_resources)
            worker.start()
            self.workers.append(worker)
            self.available.put(worker_id)
            self.conn_map[worker.parent_conn] = worker_id
        self.running = True
        self.listener_stop.clear()
        self.listener_thread = threading.Thread(target=self._listen_loop, daemon=True)
        self.listener_thread.start()

    def shutdown(self):
        if not self.running:
            return
        self.listener_stop.set()
        if self.listener_thread:
            self.listener_thread.join(timeout=1)
        with self.lock:
            tasks = list(self.tasks.items())
            self.tasks.clear()
        for _, record in tasks:
            self._finalize_instance_handle(record)
        for worker in self.workers:
            worker.stop()
        self.workers = []
        self.conn_map = {}
        self.running = False

    def dispatch_connection(self, instance, conn: socket.socket, addr, handle) -> bool:
        try:
            worker_id = self.available.get_nowait()
        except queue.Empty:
            return False
        worker = self.workers[worker_id]
        task_id = next(self.task_seq)
        server_payload = instance.build_dispatch_payload()
        handler_timeout = getattr(instance, 'handler_timeout', None)
        if handler_timeout is not None and handler_timeout <= 0:
            handler_timeout = None
        record = TaskRecord(
            task_id=task_id,
            worker_id=worker_id,
            instance=instance,
            handle=handle,
            handler_timeout=handler_timeout,
        )
        with self.lock:
            self.tasks[task_id] = record
        try:
            worker.send_task(task_id, conn, addr, server_payload)
        except Exception:
            with self.lock:
                self.tasks.pop(task_id, None)
            self.available.put(worker_id)
            raise
        finally:
            try:
                conn.close()
            except Exception:
                pass
        return True

    def _listen_loop(self):
        while not self.listener_stop.is_set():
            conns = [worker.parent_conn for worker in self.workers if worker.process and worker.process.is_alive()]
            if not conns:
                time.sleep(0.05)
                continue
            ready = []
            try:
                ready = mp_connection.wait(conns, timeout=0.1)
            except Exception:
                continue
            for conn in ready:
                worker_id = self._resolve_worker_id(conn)
                try:
                    message = conn.recv()
                except EOFError:
                    if conn in self.conn_map:
                        self.conn_map.pop(conn, None)
                    self._handle_worker_exit(worker_id)
                    continue
                cmd = message.get('cmd')
                task_id = message.get('task_id')
                if cmd == 'DONE':
                    self._complete_task(task_id)
                elif cmd == 'ERROR':
                    self._complete_task(task_id, error=message.get('error'))
                elif cmd == 'TIMEOUT':
                    self._handle_timeout_notification(task_id)

    def _complete_task(self, task_id, error=None):
        with self.lock:
            record = self.tasks.pop(task_id, None)
        if not record:
            return
        self._finalize_instance_handle(record)
        self.available.put(record.worker_id)

    def _handle_timeout_notification(self, task_id):
        with self.lock:
            record = self.tasks.pop(task_id, None)
        if not record:
            return
        metrics_module.record_timeout('handler')
        self._finalize_instance_handle(record)
        # Worker will exit via watchdog; restart handled in _handle_worker_exit.

    def _handle_worker_exit(self, worker_id):
        if worker_id is None or not self.running:
            return
        pending = []
        with self.lock:
            for task_id, record in list(self.tasks.items()):
                if record.worker_id == worker_id:
                    pending.append(record)
                    self.tasks.pop(task_id, None)
        for record in pending:
            self._finalize_instance_handle(record)
        self._restart_worker(worker_id)
        self.available.put(worker_id)

    def _restart_worker(self, worker_id):
        worker = self.workers[worker_id]
        old_conn = worker.parent_conn
        worker.restart()
        self.conn_map.pop(old_conn, None)
        self.conn_map[worker.parent_conn] = worker_id

    def _finalize_instance_handle(self, record: TaskRecord):
        try:
            record.instance._finalize_handle(record.handle)
        except Exception:
            pass

    def _resolve_worker_id(self, conn):
        worker_id = self.conn_map.get(conn)
        if worker_id is not None:
            return worker_id
        for worker in self.workers:
            if worker.parent_conn == conn:
                return worker.worker_id
        return None


def _worker_main(worker_id, control_conn, snapshot, shared_resources):
    from . import core as core_module
    from . import metrics as metrics_module
    core_module._apply_runtime_snapshot(snapshot)
    core_module._import_global_resources(shared_resources)
    try:
        metrics_resource = core_module.get_global_resource('bbws.metrics')
        metrics_module.attach_store(metrics_resource['store'], metrics_resource.get('lock'))
    except Exception:
        pass
    while True:
        message = control_conn.recv()
        cmd = message.get('cmd')
        if cmd == 'STOP':
            break
        if cmd != 'START':
            continue
        task_id = message['task_id']
        socket_info = message['socket_info']
        addr = message['addr']
        server_payload = message['server']
        try:
            conn = _receive_socket(control_conn, socket_info)
            _serve_connection(conn, addr, server_payload, core_module, control_conn, task_id)
            control_conn.send({'cmd': 'DONE', 'task_id': task_id})
        except Exception as exc:
            control_conn.send({'cmd': 'ERROR', 'task_id': task_id, 'error': str(exc)})


def _receive_socket(control_conn, socket_info):
    handle = recv_handle(control_conn)
    family = socket_info['family']
    sock_type = socket_info['type']
    proto = socket_info['proto']
    dup = socket.socket(family, sock_type, proto, fileno=handle)
    return dup


def _prepare_connection(connection, payload):
    if payload.get('ssl_enabled'):
        return _wrap_ssl_connection(connection, payload)
    keep_alive = payload.get('keep_alive_timeout')
    if keep_alive:
        try:
            connection.settimeout(keep_alive)
        except Exception:
            pass
    return connection


def _wrap_ssl_connection(conn, payload):
    cert_path = payload.get('cert_path')
    key_path = payload.get('key_path')
    handshake_timeout = payload.get('handshake_timeout') or 5
    try:
        conn.settimeout(handshake_timeout)
    except Exception:
        pass
    try:
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        if cert_path and key_path:
            context.load_cert_chain(certfile=cert_path, keyfile=key_path)
        wrapped = context.wrap_socket(conn, server_side=True)
        keep_alive = payload.get('keep_alive_timeout')
        if keep_alive:
            try:
                wrapped.settimeout(keep_alive)
            except Exception:
                pass
        return wrapped
    except Exception:
        try:
            conn.close()
        except Exception:
            pass
        return None


def _serve_connection(connection, addr, server_payload, core_module, control_conn, task_id):
    instance = _ServerContext(server_payload)
    prepared = _prepare_connection(connection, server_payload)
    if prepared is None:
        return
    timeout = server_payload.get('handler_timeout')
    controller = _TimeoutController(prepared, control_conn, task_id, timeout)
    worker_state = _WorkerState(instance, handler_signal=controller)
    try:
        core_module.servlet(prepared, addr, worker_state, server_instance=instance)
    finally:
        controller.stop()
        try:
            prepared.close()
        except Exception:
            pass


class _WorkerState:
    def __init__(self, server_instance, handler_signal=None):
        self._event = threading.Event()
        self._event.set()
        self.request_count = 0
        self.server_instance = server_instance
        self.handler_signal = handler_signal

    def is_set(self):
        return self._event.is_set()

    def set(self):
        self._event.set()

    def clear(self):
        self._event.clear()


class _ServerContext:
    def __init__(self, payload):
        self.__dict__.update(payload)

    def build_dispatch_payload(self):
        return self.__dict__


class _TimeoutController:
    def __init__(self, connection, control_conn, task_id, timeout):
        self.connection = connection
        self.control_conn = control_conn
        self.task_id = task_id
        self.timeout = timeout if timeout and timeout > 0 else None
        self._finished = threading.Event()
        self._thread = None
        self._lock = threading.Lock()

    def start(self):
        if not self.timeout:
            return
        with self._lock:
            if self._thread:
                return
            self._thread = threading.Thread(target=self._watchdog, daemon=True)
            self._thread.start()

    def end(self):
        if not self.timeout:
            return
        self._finished.set()

    def stop(self):
        if not self.timeout:
            return
        self._finished.set()
        thread = self._thread
        if thread and thread.is_alive():
            thread.join(timeout=0.05)

    def _watchdog(self):
        if self._finished.wait(self.timeout):
            return
        self._trigger_timeout()

    def _trigger_timeout(self):
        response = (
            b'HTTP/1.1 504 Gateway Timeout\r\n'
            b'Connection: close\r\n'
            b'Content-Length: 0\r\n\r\n'
        )
        try:
            self.connection.sendall(response)
        except Exception:
            pass
        try:
            self.connection.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        try:
            self.connection.close()
        except Exception:
            pass
        try:
            self.control_conn.send({'cmd': 'TIMEOUT', 'task_id': self.task_id})
        except Exception:
            pass
        os._exit(1)



from typing import Type, Union, Optional, Literal, get_type_hints, List, Dict, Any
from types import FunctionType


class ValidationError:
    """Represents a single validation error found during schema checking."""
    def __init__(self, message: str, path: List[Union[str, int]], value: Any, expected_type: Any):
        self.message = message
        self.path = path if path is not None else []
        self.value = value
        self.expected_type = expected_type

    def __str__(self):
        path_str = ".".join(map(str, self.path)) if self.path else "root"
        return f"Validation Error at '{path_str}': {self.message}"

    def __repr__(self):
        return (f"ValidationError(message='{self.message}', path={self.path!r}, "
                f"value={self.value!r}, expected_type={self.expected_type!r})")

class MultipleValidationErrors(Exception):
    """
    Exception raised when multiple validation errors occur.
    Contains a list of ValidationError objects.
    """
    def __init__(self, errors: List[ValidationError]):
        self.errors = errors
        error_messages = "\n".join(str(err) for err in errors)
        super().__init__(f"Multiple validation errors occurred:\n{error_messages}")

# --- Metaklassen ---
class SchemeMeta(type):
    def __getitem__(cls, args):
        return cls(args)

class Scheme(type, metaclass=SchemeMeta):
    '''
    Der Scheme type erlaubt es, in einem definierten JSON Schema weitere JSON Schemata
    als Typ eines Attributs festzulegen.
    '''
    def __init__(self, schema: dict):
        self.schema = schema
        
    def __str__(self):
        return f'Schema: {str(self.schema)}'
    
    def __new__(cls,*args, **kwargs):
        klass = super().__new__(cls, "Scheme", (), {})
        return klass

class ConstraintMeta(type):
    def __getitem__(cls, args):
        if len(args) == 4:
            constraint_type, operator, value, error_message = args
        elif len(args) == 3:
            constraint_type, operator, value = args
            error_message = None
        else:
            raise TypeError("Constraint expects 3 or 4 arguments: (constraint_type, operator, value, [error_message])")
        return cls(constraint_type, operator, value, error_message)

class Constraint(type, metaclass= ConstraintMeta):
    '''
    Der Constraint type erlaubt es, in Typdeklarationen die Werte, die Typen annehmen können,
    einzuschränken.
    '''
    def __init__(self, constraint_type: Type, operator: str, value, error_message: Optional[str] = None):
        self.constraint_type = constraint_type
        self.operator = operator
        self.value = value
        self.error_message = error_message

    def is_satisfied(self, value):
        temp_errors = []
        if not satisfies(value, self.constraint_type, _errors=temp_errors):
            return False
        
        is_iterable = hasattr(value, '__len__') #and not isinstance(value, (str, bytes, bytearray))

        to_compare = len(value) if is_iterable else value

        if self.operator == 'f':
            return self.value(to_compare)
        
        if self.operator == '<':
            return to_compare < self.value
        elif self.operator == '<=':
            return to_compare <= self.value
        elif self.operator == '>':
            return to_compare > self.value
        elif self.operator == '>=':
            return to_compare >= self.value
        elif self.operator == '==':
            return to_compare == self.value
        elif self.operator == '!=':
            return to_compare != self.value
        else:
            return False
        
    def __str__(self):
        op_map = {
            '<': 'less than',
            '<=': 'less than or equal to',
            '>': 'greater than',
            '>=': 'greater than or equal to',
            '==': 'equal to',
            '!=': 'not equal to',
            'f': 'satisfies custom function'
        }
        operator_str = op_map.get(self.operator, f"'{self.operator}'")

        # Use _type_to_str for consistent deep representation of the constraint_type
        type_str = _type_to_str(self.constraint_type)

        if self.operator == 'f':
            s = f"type {type_str} that {operator_str}"
            if isinstance(self.value, FunctionType):
                s += f" ({self.value.__name__ if self.value.__name__ != '<lambda>' else 'lambda'})"
            else:
                s += f" ({self.value!r})"
        elif isinstance(self.value, type):
            s = f"type {type_str} {operator_str} {self.value.__name__}"
        else:
            s = f"type {type_str} {operator_str} {self.value!r}"

        if self.error_message:
            s += f" (Error: '{self.error_message}')"
        return f"Constraint[{s}]"

    def __repr__(self):
        if self.error_message:
            return f"Constraint[{self.constraint_type!r}, '{self.operator}', {self.value!r}, {self.error_message!r}]"
        return f"Constraint[{self.constraint_type!r}, '{self.operator}', {self.value!r}]"
    
    def __new__(cls,*args, **kwargs):
        klass = super().__new__(cls, "Constraint", (), {})
        return klass

# --- Hilfsfunktionen ---
def _type_to_str(type_hint) -> str:
    """
    Recursively converts a type hint to a human-readable string representation,
    handling custom Scheme and Constraint types gracefully.
    """
    if isinstance(type_hint, (Scheme, Constraint)):
        return str(type_hint)
    
    if hasattr(type_hint, "__origin__"):
        origin = type_hint.__origin__
        args = type_hint.__args__

        if origin is Union:
            return "Union[" + ", ".join(_type_to_str(arg) for arg in args) + "]"
        elif origin is list:
            return "List[" + _type_to_str(args[0]) + "]"
        elif origin is Optional:
            return "Optional[" + _type_to_str(args[0]) + "]"
        elif origin is dict:
            return "Dict[" + _type_to_str(args[0]) + ", " + _type_to_str(args[1]) + "]"
        elif origin is Literal:
            return "Literal[" + ", ".join(repr(arg) for arg in args) + "]"
        else:
            # Fallback for other typing generics
            return f"{origin.__name__}[{', '.join(_type_to_str(arg) for arg in args)}]"
    
    # Handle built-in types and Type objects
    if isinstance(type_hint, Type):
        if type_hint.__module__ == 'builtins':
            return type_hint.__name__
        return str(type_hint)
    elif type_hint is type(None): 
        return "None"
    
    return str(type_hint) 

def satisfies(value, form, strict=False, _errors: Optional[List[ValidationError]] = None, _path: Optional[List[Union[str, int]]] = None) -> bool:
    '''
    Die satisfies Funktion erlaubt es, Werte gegen Typdeklarationen zu prüfen.
    Sammelt Validierungsfehler in `_errors` statt sofort eine Exception zu werfen.
    Gibt True zurück, wenn der Wert das Formular erfüllt, False sonst.
    Fehler werden in die übergebene `_errors` Liste akkumuliert.
    '''
    if _errors is None:
        _errors = [] 
    if _path is None:
        _path = []

    if isinstance(form, Type) and not isinstance(form, (Constraint, Scheme)):
        is_satisfied = isinstance(value, form)
        if not is_satisfied:
            _errors.append(ValidationError(
                f"Value '{value!r}' of type '{type(value).__name__}' does not match expected type '{_type_to_str(form)}'.",
                _path, value, form
            ))
        return is_satisfied
    
    
    if isinstance(form, Constraint):
        is_satisfied = form.is_satisfied(value)
        if not is_satisfied:
            error_msg = f"Value '{value!r}' does not satisfy constraint '{_type_to_str(form)}'."
            if form.error_message:
                error_msg = form.error_message.format(arg_value=value, expected_type=_type_to_str(form))
            _errors.append(ValidationError(error_msg, _path, value, form))
        return is_satisfied
    

    if hasattr(form, "__origin__"):
        origin = form.__origin__
        args = form.__args__

        if origin is Union:
            union_branch_errors = []
            if any(satisfies(value, t, strict=strict, _errors=union_branch_errors, _path=_path) for t in args):
                return True 
            else:
                _errors.append(ValidationError(
                    f"Value '{value!r}' of type '{type(value).__name__}' does not satisfy any type in Union[{', '.join(_type_to_str(arg) for arg in args)}].",
                    _path, value, form
                ))
                return False
        elif origin is list:
            if not isinstance(value, list):
                _errors.append(ValidationError(
                    f"Expected list, got '{type(value).__name__}'.",
                    _path, value, form
                ))
                return False
            inner_type = args[0]
            if not value and (getattr(inner_type, '__origin__', None) is Optional or form is Optional[List[inner_type]]):
                return True

            all_items_satisfied = True
            for i, v in enumerate(value):
                if not satisfies(v, inner_type, strict=strict, _errors=_errors, _path=_path + [i]):
                    all_items_satisfied = False
            return all_items_satisfied
        elif origin is Optional:
            inner_type = args[0]
            if value is None:
                return True 
            return satisfies(value, inner_type, strict=strict, _errors=_errors, _path=_path)
        elif origin is dict:
            key_type, value_type = args
            if not isinstance(value, dict):
                _errors.append(ValidationError(
                    f"Expected dict, got '{type(value).__name__}'.",
                    _path, value, form
                ))
                return False
            all_kv_satisfied = True
            for k, v in value.items():
                if not satisfies(k, key_type, strict=strict, _errors=_errors, _path=_path + [f"key({k!r})"]):
                    all_kv_satisfied = False
                if not satisfies(v, value_type, strict=strict, _errors=_errors, _path=_path + [k]):
                    all_kv_satisfied = False
            return all_kv_satisfied
        elif origin is Literal:
            is_satisfied = value in args
            if not is_satisfied:
                _errors.append(ValidationError(
                    f"Value '{value!r}' is not one of the allowed literals: {', '.join(repr(arg) for arg in args)}.",
                    _path, value, form
                ))
            return is_satisfied
        else:
            _errors.append(ValidationError(
                f"Unsupported typing generic origin: '{origin.__name__}'.",
                _path, value, form
            ))
            return False 
    elif isinstance(form, Scheme):
        if not isinstance(value, dict):
            _errors.append(ValidationError(
                f"Expected dictionary for Scheme, got '{type(value).__name__}'.",
                _path, value, form
            ))
            return False
        original_error_count = len(_errors)
        validated_sub_instance = validate(value, form.schema, strict=strict, _errors=_errors, _path=_path)
        if len(_errors) > original_error_count:
            return False
        value.update(validated_sub_instance)
        return True

    elif form is type(None):
        is_satisfied = (value is None)
        if not is_satisfied:
            _errors.append(ValidationError(
                f"Expected None, got '{type(value).__name__}'.",
                _path, value, form
            ))
        return is_satisfied
    else:
        _errors.append(ValidationError(
            f"Unsupported type form or value: '{_type_to_str(form)}'.",
            _path, value, form
        ))
        return False

def validate(instance: dict, form: dict, strict: bool, _errors: Optional[List[ValidationError]] = None, _path: Optional[List[Union[str, int]]] = None) -> dict:
    '''
    Validiert eine Instanz (dict) gegen ein Schema (dict).
    Sammelt alle Validierungsfehler in `_errors` und gibt ein (partiell) validiertes Dictionary zurück.
    '''
    if _errors is None:
        _errors = []
    if _path is None:
        _path = []

    if not isinstance(instance, dict):
        _errors.append(ValidationError(f"Instance must be a dictionary, got {type(instance).__name__}.", _path, instance, dict))
        return {} 

    validated_instance = {}
    
    for key, schema_val in form.items():
        current_path = _path + [key]
        if not isinstance(schema_val, dict) or 'type' not in schema_val:
            continue 
        
        value = instance.get(key)
        field_type = schema_val['type']
        default = schema_val.get('default')

        key_is_present = key in instance

        if key_is_present:
            satisfies(value, field_type, strict=strict, _errors=_errors, _path=current_path)
            validated_instance[key] = value 
        elif default is not None:
            validated_instance[key] = default
        elif strict:
            if getattr(field_type, '__origin__', None) is not Optional:
                _errors.append(ValidationError(
                    f"Strict mode: Missing required field '{key}' and no default value provided.",
                    current_path, None, field_type
                ))
    if strict:
        for key in instance.keys():
            if key not in form:
                _errors.append(ValidationError(
                    f"Strict mode: Field '{key}' is not defined in the schema.",
                    _path + [key], instance[key], None 
                ))

    return validated_instance


def typed(func):
    '''
    Checks argument and return types at runtime.
    Raises MultipleValidationErrors if any checks fail, containing all found errors.
    '''
    def wrapper(*args, **kwargs):
        type_hints = get_type_hints(func)
        func_params = list(func.__code__.co_varnames[:func.__code__.co_argcount])
        
        bound_args = {}
        for i, arg_value in enumerate(args):
            if i < len(func_params):
                bound_args[func_params[i]] = arg_value
        bound_args.update(kwargs)

        all_errors: List[ValidationError] = []
        for arg_name, expected_type in type_hints.items():
            if arg_name == 'return':
                continue

            if arg_name in bound_args:
                arg_value = bound_args[arg_name]
                satisfies(arg_value, expected_type, _errors=all_errors, _path=[arg_name])
            elif arg_name in func_params and arg_name not in bound_args:
                if getattr(expected_type, '__origin__', None) is not Optional:
                    all_errors.append(ValidationError(
                        f"Missing required argument: '{arg_name}'.",
                        [arg_name], None, expected_type
                    ))
        if all_errors:
            raise MultipleValidationErrors(all_errors)
        result = func(*args, **kwargs)
        expected_return_type = type_hints.get('return')
        if expected_return_type is not None:
            satisfies(result, expected_return_type, _errors=all_errors, _path=["<return_value>"])
        if all_errors:
            raise MultipleValidationErrors(all_errors)
            
        return result
    
    return wrapper



HEX_CHARS = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']


#TODO: unefficient af fix that 

def leading_ones_count(encoded_byte):
    byte_value = int(encoded_byte, 16)
    count = 0
    mask = 0x80 
    while mask != 0 and byte_value & mask:
        count += 1
        mask >>= 1
    return count

def unescape_url(url:str) -> str:
    unescaped = ''
    escape_sequence = ''
    index = 0
    count = 0
    url_len = len(url)
    
    while index < url_len:
        
        if len(escape_sequence) == 0 and url[index] == '%' and index + 3 <= url_len and url[index+1] in HEX_CHARS and url[index+2] in HEX_CHARS:
            escape_byte = url[index+1:index+3]
            count =  leading_ones_count(escape_byte)
            escape_sequence += escape_byte
            index += 3
            
        elif url[index] == '%' and index + 3 <= url_len and url[index+1] in HEX_CHARS and url[index+2] in HEX_CHARS and count > 1:
            escape_byte = url[index+1:index+3]
            count -= 1
            escape_sequence+= escape_byte
            index += 3
            
        elif escape_sequence:
            unescaped += decode_hex_string(escape_sequence)
            escape_sequence = ''
            count = 0
        else:
            unescaped += url[index]
            index += 1

    return unescaped if not escape_sequence else unescaped + decode_hex_string(escape_sequence)
            

def decode_hex_string(hex_string:str) ->str:
    try:
        decoded_bytes = bytes.fromhex(hex_string)
        unicode_char = decoded_bytes.decode('utf-8')
        return unicode_char
    except UnicodeDecodeError:
        return ''


def _normalize_port(port_str: str | None):
    if port_str is None:
        return None
    if port_str == '':
        return None
    if not port_str.isdigit():
        raise ValueError(f"Invalid port specifier '{port_str}'.")
    return int(port_str)


def _parse_ip_port(component: str):
    component = component.strip()
    if not component:
        return (None, None)
    ip = None
    port = None
    if component.startswith('['):
        end = component.find(']')
        if end == -1:
            raise ValueError(f"Invalid IPv6 specifier '{component}'.")
        ip = component[1:end]
        remainder = component[end+1:]
        if remainder.startswith(':'):
            port = _normalize_port(remainder[1:])
        elif remainder:
            raise ValueError(f"Invalid specifier '{component}'.")
    else:
        if component.startswith(':'):
            port = _normalize_port(component[1:])
        elif component.count(':') == 1:
            ip_part, port_part = component.split(':', 1)
            ip = ip_part or None
            port = _normalize_port(port_part)
        else:
            ip = component or None
    return (ip, port)


def split_route_scope(route):
    if isinstance(route, UrlTemplate):
        return (route.ip, route.port, route.domain, route.path_template)
    if not isinstance(route, str):
        return (None, None, None, route)
    slash_index = route.find('/')
    if slash_index == -1:
        prefix = route
        path = route if route.startswith('/') else '/' + route
    else:
        prefix = route[:slash_index]
        path = route[slash_index:]
    ip = port = domain = None
    if '::' in prefix:
        ip_part, domain_part = prefix.rsplit('::', 1)
        ip, port = _parse_ip_port(ip_part)
        domain = domain_part.rstrip(':') or None
    elif prefix:
        ip, port = _parse_ip_port(prefix)
    if not path.startswith('/'):
        path = '/' + path if path else '/'
    return (ip, port, domain, path)


def format_ip_port(ip, port):
    if ip is None and port is None:
        return ''
    ip_repr = ''
    if ip is not None:
        ip_repr = f'[{ip}]' if (':' in ip and not ip.startswith('[')) else ip
    if port is not None:
        port_str = str(port)
        if ip_repr:
            return f'{ip_repr}:{port_str}'
        return f':{port_str}'
    return ip_repr


def _is_decimal(value: str) -> bool:
    if value.count('.') != 1:
        return False
    left, right = value.split('.', 1)
    return bool(left) and bool(right) and left.isdigit() and right.isdigit()


class UrlTemplate:
    """
    A URL template wraps a route string with optional IP/port/domain selectors
    plus path placeholders.

    Syntax:
        ip:port::domain:/path/to/{placeholders}

    Examples:
        - '::/status'                                -> matches every endpoint regardless of IP or domain
        - '127.0.0.1::/debug'                        -> IPv4 127.0.0.1, all ports, all domains
        - ':8443::/metrics'                          -> any IP, port 8443
        - ':::example.com:/domain-only'              -> all IPs/ports, only domain `example.com`
        - '127.0.0.1:8000::example.com:/foo'         -> specific IP/port/domain combination
        - '[::1]:8080::/ipv6/{path:path}'            -> IPv6 loopback, port 8080, path wildcard

    Path placeholders use the familiar '{name:type}' syntax and support the built-in
    types ('int', 'str', 'float', 'bool', 'path').
    """

    _PLACEHOLDER_TYPES = {
        'int': (
            lambda value: value.isdigit(),
            int,
        ),
        'float': (
            lambda value: _is_decimal(value),
            float,
        ),
        'bool': (
            lambda value: value in ('True', 'False'),
            lambda value: value == 'True',
        ),
        'str': (
            lambda value: bool(value) and all(ch.isalnum() or ch == '_' for ch in value),
            lambda value: value,
        ),
        'path': (
            lambda value: bool(value),
            lambda value: value,
        ),
    }

    def __init__(self, template_string):
        self.template = template_string
        self.ip, self.port, self.domain, self.path_template = split_route_scope(template_string)
        self.handler = None
        self.type = None
        self._is_root = self.path_template == '/'
        self._requires_trailing_slash = bool(self.path_template.endswith('/') and not self._is_root)
        self._segments = self._compile_segments()
        self._path_wildcard_index = None
        self._path_placeholder_name = None
        for idx, segment in enumerate(self._segments):
            if segment.get('path_placeholder'):
                self._path_wildcard_index = idx
                self._path_placeholder_name = segment['path_placeholder']
                break

    def _compile_segments(self):
        if self._is_root:
            return []
        raw = self.path_template.strip('/')
        if not raw:
            return []
        parts = raw.split('/')
        segments = []
        path_placeholder_seen = False
        for index, part in enumerate(parts):
            tokens = self._parse_segment_tokens(part)
            if tokens and any(tok['kind'] == 'var' and tok['var_type'] == 'path' for tok in tokens):
                if path_placeholder_seen:
                    raise ValueError('Only one {path:path} placeholder is allowed per UrlTemplate.')
                if len(tokens) != 1 or tokens[0]['var_type'] != 'path':
                    raise ValueError('{path:path} must occupy an entire path segment.')
                if index != len(parts) - 1:
                    raise ValueError('{path:path} must be the final segment in a UrlTemplate.')
                path_placeholder_seen = True
                segments.append({
                    'tokens': tokens,
                    'path_placeholder': tokens[0]['name'],
                })
                continue
            segments.append({'tokens': tokens})
        return segments

    def _parse_segment_tokens(self, segment):
        if not segment:
            return []
        tokens = []
        idx = 0
        last_token_was_var = False
        while idx < len(segment):
            start = segment.find('{', idx)
            if start == -1:
                literal = segment[idx:]
                if literal:
                    tokens.append({'kind': 'literal', 'value': literal})
                break
            if start > idx:
                literal = segment[idx:start]
                if literal:
                    tokens.append({'kind': 'literal', 'value': literal})
                    last_token_was_var = False
            end = segment.find('}', start)
            if end == -1:
                raise ValueError(f"Malformed placeholder in '{segment}'.")
            inner = segment[start + 1:end]
            if ':' not in inner:
                raise ValueError(f"Invalid placeholder '{inner}', expected syntax {{name:type}}.")
            name, type_name = inner.split(':', 1)
            if not name or not all(ch.isalnum() or ch == '_' for ch in name):
                raise ValueError(f"Invalid placeholder name '{name}'.")
            if type_name not in self._PLACEHOLDER_TYPES:
                raise ValueError(f"Unknown placeholder type '{type_name}'.")
            if last_token_was_var:
                raise ValueError('Adjacent placeholders must be separated by literal text.')
            tokens.append({'kind': 'var', 'name': name, 'var_type': type_name})
            last_token_was_var = True
            idx = end + 1
        return tokens

    def _split_request_segments(self, path):
        if path == '/':
            return []
        trimmed = path.strip('/')
        if not trimmed:
            return []
        return trimmed.split('/')

    def _next_literal(self, tokens, start_index):
        for offset in range(start_index, len(tokens)):
            token = tokens[offset]
            if token['kind'] == 'literal' and token['value']:
                return token['value']
        return None

    def _match_segment(self, tokens, segment):
        if not tokens:
            return {} if segment == '' else None
        position = 0
        values = {}
        for idx, token in enumerate(tokens):
            if token['kind'] == 'literal':
                literal = token['value']
                if not segment.startswith(literal, position):
                    return None
                position += len(literal)
            else:
                var_type = token['var_type']
                validator, converter = self._PLACEHOLDER_TYPES[var_type]
                next_literal = self._next_literal(tokens, idx + 1)
                if next_literal is None:
                    value = segment[position:]
                    position = len(segment)
                else:
                    next_index = segment.find(next_literal, position)
                    if next_index == -1:
                        return None
                    value = segment[position:next_index]
                    position = next_index
                if not value or not validator(value):
                    return None
                values[token['name']] = converter(value)
        if position != len(segment):
            return None
        return values

    def _match_path(self, path, ip=None, port=None, domain=None):
        if not path or not path.startswith('/'):
            return None
        if self.ip is not None and self.ip != ip:
            return None
        if self.port is not None and self.port != port:
            return None
        if self.domain:
            if domain is None or self.domain != domain:
                return None

        if self._is_root:
            return {} if path == '/' else None

        has_trailing_slash = path.endswith('/') and path != '/'
        if self._requires_trailing_slash != has_trailing_slash:
            return None

        request_segments = self._split_request_segments(path)
        extracted = {}

        if self._path_wildcard_index is not None:
            min_segments = self._path_wildcard_index + 1
            if len(request_segments) < min_segments:
                return None
        elif len(request_segments) != len(self._segments):
            return None

        for index, segment_def in enumerate(self._segments):
            if self._path_wildcard_index is not None and index == self._path_wildcard_index:
                remaining = request_segments[index:]
                value = '/'.join(remaining)
                if not value:
                    return None
                validator, converter = self._PLACEHOLDER_TYPES['path']
                if not validator(value):
                    return None
                extracted[segment_def['path_placeholder']] = converter(value)
                break
            segment_value = request_segments[index]
            segment_result = self._match_segment(segment_def['tokens'], segment_value)
            if segment_result is None:
                return None
            extracted.update(segment_result)

        return extracted

    def extract(self, url):
        return self._match_path(url)

    def matches(self, ip, port, domain, path):
        return self._match_path(path, ip, port, domain) is not None

    def __eq__(self, other):
        if isinstance(other, str):
            return self._match_path(other) is not None
        if isinstance(other, self.__class__):
            return (
                self.path_template == other.path_template
                and self.ip == other.ip
                and self.port == other.port
                and self.domain == other.domain
            )
        return False



import sys as system
import json
import os
from typing import List, Literal, Optional, Union

from .__init__ import MAIN_PATH
from .type_validation import Constraint, Scheme, ValidationError, satisfies

ERROR_PREFIX = '[CONFIG_ERROR]'

HOST_ENTRY_SCHEMA = {
    "host": {"type": str},
    "cert_path": {"type": str},
    "key_path": {"type": str},
}

HOST_TYPE = Union[str, List[Scheme[HOST_ENTRY_SCHEMA]]]

SERVER_BASE_SCHEMA = {
    "ip": {"type": str},
    "port": {"type": Constraint[int, '>=', 0]},
    "queue_size": {"type": Constraint[int, '>=', 0]},
    "max_threads": {"type": Optional[Constraint[int, '>=', 0]], "default": None},
    "keep_alive_timeout": {"type": Optional[Constraint[int, '>=', 0]], "default": None},
    "keep_alive_max_requests": {"type": Optional[Constraint[int, '>=', 0]], "default": None},
    "header_timeout": {"type": Optional[Constraint[int, '>=', 0]], "default": None},
    "body_min_rate_bytes_per_sec": {"type": Optional[Constraint[int, '>=', 0]], "default": None},
    "handler_timeout": {"type": Optional[Constraint[int, '>=', 0]], "default": None},
    "ssl_handshake_timeout": {"type": Optional[Constraint[int, '>=', 0]], "default": None},
    "https-redirect": {"type": Optional[Literal[True, False]], "default": False},
    "https-redirect-escape-paths": {"type": Optional[List[str]], "default": []},
    "update-cert-state": {"type": Optional[Literal[True, False]], "default": False},
}

SSL_DISABLED_SCHEMA = {
    **SERVER_BASE_SCHEMA,
    "SSL": {"type": Literal[False]},
    "host": {"type": HOST_TYPE, "default": ""},
    "cert_path": {"type": Optional[str], "default": ""},
    "key_path": {"type": Optional[str], "default": ""},
}

SSL_ENABLED_SCHEMA = {
    **SERVER_BASE_SCHEMA,
    "SSL": {"type": Literal[True]},
    "host": {"type": HOST_TYPE, "default": ""},
    "cert_path": {"type": Optional[str], "default": ""},
    "key_path": {"type": Optional[str], "default": ""},
}

SERVER_SCHEMA = Union[Scheme[SSL_DISABLED_SCHEMA], Scheme[SSL_ENABLED_SCHEMA]]

DEFAULT_WORKER_PROCESSES = max(1, os.cpu_count() or 1)

CONFIG_SCHEMA = Scheme({
    "max_threads": {"type": Constraint[int, '>=', 0], "default": 100},
    "max_header_size": {"type": Constraint[int, '>', 0], "default": 16384},
    "max_body_size": {"type": Constraint[int, '>', 0], "default": 10485760},
    "max_url_length": {"type": Constraint[int, '>', 0], "default": 2048},
    "keep_alive_timeout": {"type": Constraint[int, '>=', 0], "default": 15},
    "keep_alive_max_requests": {"type": Constraint[int, '>=', 0], "default": 100},
    "header_timeout": {"type": Constraint[int, '>=', 0], "default": 10},
    "body_min_rate_bytes_per_sec": {"type": Constraint[int, '>=', 0], "default": 1024},
    "handler_timeout": {"type": Constraint[int, '>=', 0], "default": 30},
    "ssl_handshake_timeout": {"type": Constraint[int, '>=', 0], "default": 5},
    "worker_processes": {"type": Constraint[int, '>=', 1], "default": DEFAULT_WORKER_PROCESSES},
    "server": {"type": List[SERVER_SCHEMA]},
})


def _normalize_host_entries(host):
    if isinstance(host, list):
        return host
    if isinstance(host, str) and host:
        return [{"host": host, "cert_path": "", "key_path": ""}]
    return []


class Config:
    def __init__(self) -> None:
        """Load and validate configuration values from config.json."""

        config_path = MAIN_PATH + "/config/config.json"

        with open(config_path, 'r', encoding='utf-8') as file:
            config_raw = json.load(file)

        config_data = dict(config_raw)
        server_timeout_keys = [
            'keep_alive_timeout',
            'keep_alive_max_requests',
            'header_timeout',
            'body_min_rate_bytes_per_sec',
            'handler_timeout',
            'ssl_handshake_timeout',
        ]
        for entry in config_data.get('server', []):
            for key in server_timeout_keys:
                entry.setdefault(key, None)
        errors: List[ValidationError] = []
        is_valid = satisfies(config_data, CONFIG_SCHEMA, strict=True, _errors=errors)
        if not is_valid:
            for error in errors:
                print(ERROR_PREFIX, str(error))
            system.exit(0)

        self.MAX_THREADS = config_data['max_threads']
        self.MAX_HEADER_SIZE = config_data['max_header_size']
        self.MAX_BODY_SIZE = config_data['max_body_size']
        self.MAX_URL_LENGTH = config_data['max_url_length']
        self.KEEP_ALIVE_TIMEOUT = config_data['keep_alive_timeout']
        self.KEEP_ALIVE_MAX_REQUESTS = config_data['keep_alive_max_requests']
        self.HEADER_TIMEOUT = config_data['header_timeout']
        self.BODY_MIN_RATE_BYTES_PER_SEC = config_data['body_min_rate_bytes_per_sec']
        self.HANDLER_TIMEOUT = config_data['handler_timeout']
        self.SSL_HANDSHAKE_TIMEOUT = config_data['ssl_handshake_timeout']
        self.WORKER_PROCESSES = config_data['worker_processes']
        self.SERVERS = []
        for entry in config_data['server']:
            normalized_entry = dict(entry)
            normalized_entry['host'] = _normalize_host_entries(normalized_entry.get('host', ''))
            if normalized_entry['SSL']:
                cert_path = normalized_entry.get('cert_path', '')
                key_path = normalized_entry.get('key_path', '')
                host_entries = normalized_entry['host']
                if not host_entries and (not cert_path or not key_path):
                    print(ERROR_PREFIX, "SSL server requires 'cert_path' and 'key_path' when no host list is provided.")
                    system.exit(0)
                if host_entries and (not cert_path or not key_path):
                    first_host = host_entries[0]
                    normalized_entry['cert_path'] = first_host['cert_path']
                    normalized_entry['key_path'] = first_host['key_path']
            else:
                normalized_entry['cert_path'] = normalized_entry.get('cert_path', '')
                normalized_entry['key_path'] = normalized_entry.get('key_path', '')
            if normalized_entry.get('max_threads') is None:
                normalized_entry['max_threads'] = self.MAX_THREADS
            normalized_entry['https-redirect'] = bool(normalized_entry.get('https-redirect', False))
            normalized_entry['https-redirect-escape-paths'] = list(normalized_entry.get('https-redirect-escape-paths', []))
            normalized_entry['update-cert-state'] = bool(normalized_entry.get('update-cert-state', False))
            timeout_defaults = {
                'keep_alive_timeout': self.KEEP_ALIVE_TIMEOUT,
                'keep_alive_max_requests': self.KEEP_ALIVE_MAX_REQUESTS,
                'header_timeout': self.HEADER_TIMEOUT,
                'body_min_rate_bytes_per_sec': self.BODY_MIN_RATE_BYTES_PER_SEC,
                'handler_timeout': self.HANDLER_TIMEOUT,
                'ssl_handshake_timeout': self.SSL_HANDSHAKE_TIMEOUT,
            }
            for key, default in timeout_defaults.items():
                if normalized_entry.get(key) is None:
                    normalized_entry[key] = default
            self.SERVERS.append(normalized_entry)

    def _get_server_value(self, key):
        if not self.SERVERS:
            return None
        return self.SERVERS[0].get(key)

    def _set_server_value(self, key, value):
        if not self.SERVERS:
            return
        self.SERVERS[0][key] = value

    @property
    def SERVER_IP(self):
        return self._get_server_value('ip')

    @SERVER_IP.setter
    def SERVER_IP(self, value):
        self._set_server_value('ip', value)

    @property
    def SERVER_PORT(self):
        return self._get_server_value('port')

    @SERVER_PORT.setter
    def SERVER_PORT(self, value):
        self._set_server_value('port', value)

    @property
    def QUE_SIZE(self):
        return self._get_server_value('queue_size')

    @QUE_SIZE.setter
    def QUE_SIZE(self, value):
        self._set_server_value('queue_size', value)

    @property
    def SSL(self):
        return self._get_server_value('SSL')

    @SSL.setter
    def SSL(self, value):
        self._set_server_value('SSL', value)

    @property
    def HOST(self):
        return self._get_server_value('host')

    @HOST.setter
    def HOST(self, value):
        self._set_server_value('host', _normalize_host_entries(value))

    @property
    def CERT_PATH(self):
        return self._get_server_value('cert_path')

    @CERT_PATH.setter
    def CERT_PATH(self, value):
        self._set_server_value('cert_path', value)

    @property
    def KEY_PATH(self):
        return self._get_server_value('key_path')

    @KEY_PATH.setter
    def KEY_PATH(self, value):
        self._set_server_value('key_path', value)

    @property
    def SERVER_MAX_THREADS(self):
        return self._get_server_value('max_threads')

    @SERVER_MAX_THREADS.setter
    def SERVER_MAX_THREADS(self, value):
        self._set_server_value('max_threads', value)

    @property
    def HTTPS_REDIRECT(self):
        return self._get_server_value('https-redirect')

    @HTTPS_REDIRECT.setter
    def HTTPS_REDIRECT(self, value):
        self._set_server_value('https-redirect', bool(value))

    @property
    def HTTPS_REDIRECT_ESCAPE_PATHS(self):
        return self._get_server_value('https-redirect-escape-paths')

    @HTTPS_REDIRECT_ESCAPE_PATHS.setter
    def HTTPS_REDIRECT_ESCAPE_PATHS(self, value):
        paths = list(value) if isinstance(value, list) else []
        self._set_server_value('https-redirect-escape-paths', paths)

    @property
    def UPDATE_CERT_STATE(self):
        return self._get_server_value('update-cert-state')

    @UPDATE_CERT_STATE.setter
    def UPDATE_CERT_STATE(self, value):
        self._set_server_value('update-cert-state', bool(value))



from __future__ import annotations

from threading import Lock
from typing import Mapping, Any

def _create_default_store():
    return {'header': 0, 'body': 0, 'handler': 0}


_TIMEOUT_COUNTERS: dict[str, int] = _create_default_store()
_COUNTER_LOCK: Lock | Any = Lock()


def attach_store(store, lock=None):
    """Attach a shared store (and optional lock) for timeout metrics."""
    global _TIMEOUT_COUNTERS, _COUNTER_LOCK
    _TIMEOUT_COUNTERS = store
    if lock is not None:
        _COUNTER_LOCK = lock


def record_timeout(kind: str) -> None:
    """Increment the timeout counter for ``kind`` if it is tracked."""
    normalized = kind.lower()
    with _COUNTER_LOCK:
        if normalized not in _TIMEOUT_COUNTERS:
            _TIMEOUT_COUNTERS[normalized] = 0
        _TIMEOUT_COUNTERS[normalized] += 1


def get_timeout_metrics() -> dict[str, int]:
    """Return a snapshot suitable for JSON serialization."""
    with _COUNTER_LOCK:
        return {f'timeouts.{key}': value for key, value in _TIMEOUT_COUNTERS.items()}


def reset_timeout_metrics(initial: Mapping[str, int] | None = None) -> None:
    """Reset all timeout counters (mainly useful for tests)."""
    with _COUNTER_LOCK:
        if initial:
            _TIMEOUT_COUNTERS.clear()
            for key, value in initial.items():
                _TIMEOUT_COUNTERS[key.lower()] = int(value)
        else:
            for key in list(_TIMEOUT_COUNTERS.keys()):
                _TIMEOUT_COUNTERS[key] = 0
